/*
 *  Copyright (C) 2011-2012 Dmitry Moskalchuk <dm@crystax.net>
 *  This file is part of Kodi - https://kodi.tv
 *
 *  SPDX-License-Identifier: BSD-2-Clause-Views
 *  See LICENSES/README.md for more information.
 */

CRYSTAX_PP_STEP(jboolean)
CRYSTAX_PP_STEP(jbyte)
CRYSTAX_PP_STEP(jchar)
CRYSTAX_PP_STEP(jshort)
CRYSTAX_PP_STEP(jint)
CRYSTAX_PP_STEP(jlong)
CRYSTAX_PP_STEP(jfloat)
CRYSTAX_PP_STEP(jdouble)
CRYSTAX_PP_STEP(jhobject)
CRYSTAX_PP_STEP(jhclass)
CRYSTAX_PP_STEP(jhstring)
CRYSTAX_PP_STEP(jhthrowable)
CRYSTAX_PP_STEP(jharray)
CRYSTAX_PP_STEP(jhbooleanArray)
CRYSTAX_PP_STEP(jhbyteArray)
CRYSTAX_PP_STEP(jhshortArray)
CRYSTAX_PP_STEP(jhintArray)
CRYSTAX_PP_STEP(jhlongArray)
CRYSTAX_PP_STEP(jhfloatArray)
CRYSTAX_PP_STEP(jhdoubleArray)
CRYSTAX_PP_STEP(jhobjectArray)
